@@ -1,6 +1,5 @@
-//#define DEBUG
 /******************************************************************************
- * Copyright 1994-2002,2003 by Thomas E. Dickey                               *
+ * Copyright 1994-2004,2005 by Thomas E. Dickey                               *
  * All Rights Reserved.                                                       *
  *                                                                            *
  * Permission to use, copy, modify, and distribute this software and its      *
@@ -21,7 +20,7 @@
  ******************************************************************************/
 
 #ifndef	NO_IDENT
-static char *Id = "$Id: diffstat.c,v 1.33 2003/02/15 00:57:22 tom Exp $";
+static const char *Id = "$Id: diffstat.c,v 1.38 2005/01/16 18:03:32 Eric.Blake Exp $";
 #endif
 
 /*
@@ -29,6 +28,23 @@ static char *Id = "$Id: diffstat.c,v 1.33 2003/02/15 00:57:22 tom Exp $";
  * Author:	T.E.Dickey
  * Created:	02 Feb 1992
  * Modified:
+ *		10 Jan 2005, add support for '--help' and '--version' (Patch
+ *			     by Eric Blake <ebb9@byu.net>.)
+ *		16 Dec 2004, fix a different case for data beginning with "--"
+ *			     which was treated as a header line.
+ *		14 Dec 2004, Fix allocation problems.  Open files in binary
+ *			     mode for reading.  Getopt returns -1, not
+ *			     necessarily EOF.  Add const where useful.  Use
+ *			     NO_IDENT where necessary.  malloc() comes from
+ *			     <stdlib.h> in standard systems (Patch by Eric
+ *			     Blake <ebb9@byu.net>.)
+ *		08 Nov 2004, minor fix for resync of unified diffs checks for
+ *			     range (line beginning with '@' without header
+ *			     lines (successive lines beginning with "---" and
+ *			     "+++").  Fix a few problems reported by valgrind.
+ *		09 Nov 2003, modify check for lines beginning with '-' or '+'
+ *			     to treat only "---" in old-style diffs as a
+ *			     special case.
  *		14 Feb 2003, modify check for filenames to allow for some cases
  *			     of incomplete dates (the reported example omitted
  *			     the day of the month).  Correct a typo in usage().
@@ -115,8 +131,6 @@ extern int isatty();
 
 #ifdef HAVE_MALLOC_H
 #include <malloc.h>
-#else
-extern char *malloc();
 #endif
 
 #ifdef HAVE_GETOPT_H
@@ -127,13 +141,6 @@ extern char *optarg;
 extern int optind;
 #endif
 
-#if !defined(TRUE) || (TRUE != 1)
-#undef  TRUE
-#undef  FALSE
-#define	TRUE		1
-#define	FALSE		0
-#endif
-
 #if !defined(EXIT_SUCCESS)
 #define EXIT_SUCCESS 0
 #define EXIT_FAILURE 1
@@ -151,6 +158,8 @@ extern int optind;
 #define EOS     '\0'
 #define BLANK   ' '
 
+#define UC(c)   ((unsigned char)(c))
+
 #ifdef DEBUG
 #define TRACE(p) printf p
 #else
@@ -194,12 +203,22 @@ static long plot_scale;		/* the effective scale (1:maximum) */
 /******************************************************************************/
 
 static void
-failed(char *s)
+failed(const char *s)
 {
     perror(s);
     exit(EXIT_FAILURE);
 }
 
+/* malloc wrapper that never returns NULL */
+static void *
+xmalloc(size_t s)
+{
+    void *p;
+    if ((p = malloc(s)) == NULL)
+	failed("malloc");
+    return p;
+}
+
 static void
 blip(int c)
 {
@@ -210,13 +229,13 @@ blip(int c)
 }
 
 static char *
-new_string(char *s)
+new_string(const char *s)
 {
-    return strcpy((char *) malloc((unsigned) (strlen(s) + 1)), s);
+    return strcpy((char *) xmalloc((size_t) (strlen(s) + 1)), s);
 }
 
 static DATA *
-new_data(char *name)
+new_data(const char *name)
 {
     DATA *p, *q, *r;
 
@@ -233,7 +252,7 @@ new_data(char *name)
 	if (sort_names && (cmp > 0))
 	    break;
     }
-    r = (DATA *) malloc(sizeof(DATA));
+    r = (DATA *) xmalloc(sizeof(DATA));
     if (q != 0)
 	q->link = r;
     else
@@ -251,8 +270,7 @@ new_data(char *name)
 }
 
 /*
- * Remove a unneeded data item from the linked list.  Don't free the name,
- * since we may want it in another context.
+ * Remove a unneeded data item from the linked list.  Free the name as well.
  */
 static void
 delink(DATA * data)
@@ -267,6 +285,7 @@ delink(DATA * data)
 		q->link = p->link;
 	    else
 		all_data = p->link;
+	    free(p->name);
 	    free(p);
 	    return;
 	}
@@ -275,23 +294,27 @@ delink(DATA * data)
 
 /* like strncmp, but without the 3rd argument */
 static int
-match(char *s, char *p)
+match(const char *s, const char *p)
 {
-    int ok = FALSE;
+    int ok = 0;
 
     while (*s != EOS) {
 	if (*p == EOS) {
-	    ok = TRUE;
+	    ok = 1;
 	    break;
 	}
 	if (*s++ != *p++)
 	    break;
+	if (*s == EOS && *p == EOS) {
+	    ok = 1;
+	    break;
+	}
     }
     return ok;
 }
 
 static int
-version_num(char *s)
+version_num(const char *s)
 {
     int main_ver, sub_ver;
     char temp[2];
@@ -302,7 +325,7 @@ version_num(char *s)
  * Check for a range of line-numbers, used in editing scripts.
  */
 static int
-edit_range(char *s)
+edit_range(const char *s)
 {
     int first, last;
     char temp[2];
@@ -316,20 +339,22 @@ edit_range(char *s)
  * shows that the numbers are a line-number followed by a count.
  */
 static int
-decode_range(char *s, int *first, int *second)
+decode_range(const char *s, int *first, int *second)
 {
     char check;
     if (sscanf(s, "%d,%d%c", first, second, &check) == 2) {
+	TRACE(("decode_range #1 first=%d, second=%d\n", *first, *second));
 	return 1;
     } else if (sscanf(s, "%d%c", first, &check) == 1) {
 	*second = *first;	/* diffutils 2.7 does this */
+	TRACE(("decode_range #2 first=%d, second=%d\n", *first, *second));
 	return 1;
     }
     return 0;
 }
 
 static int
-HadDiffs(DATA * data)
+HadDiffs(const DATA * data)
 {
     return data->ins != 0
 	|| data->del != 0
@@ -341,25 +366,25 @@ HadDiffs(DATA * data)
  * If the given path is not one of the "ignore" paths, then return true.
  */
 static int
-can_be_merged(char *path)
+can_be_merged(const char *path)
 {
     if (strcmp(path, "")
 	&& strcmp(path, "/dev/null")
 	&& strncmp(path, "/tmp/", 5))
-	return TRUE;
-    return FALSE;
+	return 1;
+    return 0;
 }
 
 static int
-is_leaf(char *leaf, char *path)
+is_leaf(const char *leaf, const char *path)
 {
     char *s;
 
     if (strchr(leaf, PATHSEP) == 0
 	&& (s = strrchr(path, PATHSEP)) != 0
 	&& !strcmp(++s, leaf))
-	return TRUE;
-    return FALSE;
+	return 1;
+    return 0;
 }
 
 static char *
@@ -427,7 +452,7 @@ do_merging(DATA * data, char *path)
 }
 
 static int
-begin_data(DATA * p)
+begin_data(const DATA * p)
 {
     if (!can_be_merged(p->name)
 	&& strchr(p->name, PATHSEP) != 0) {
@@ -441,7 +466,7 @@ begin_data(DATA * p)
 static char *
 skip_blanks(char *s)
 {
-    while (isspace(*s))
+    while (isspace(UC(*s)))
 	++s;
     return s;
 }
@@ -452,7 +477,7 @@ skip_options(char *params)
     while (*params != '\0') {
 	params = skip_blanks(params);
 	if (*params == '-') {
-	    while (isgraph(*params))
+	    while (isgraph(UC(*params)))
 		params++;
 	} else {
 	    break;
@@ -468,9 +493,12 @@ static void
 dequote(char *s)
 {
     int len = strlen(s);
+    int n;
 
     if (*s == SQUOTE && len > 2 && s[len - 1] == SQUOTE) {
-	strcpy(s, s + 1);
+	for (n = 0; (s[n] = s[n + 1]) != EOS; ++n) {
+	    ;
+	}
 	s[len - 2] = EOS;
     }
 }
@@ -481,8 +509,7 @@ dequote(char *s)
 static void
 fixed_buffer(char **buffer, size_t want)
 {
-    if ((*buffer = malloc(want)) == 0)
-	failed("malloc");
+    *buffer = xmalloc(want);
 }
 
 /*
@@ -509,7 +536,7 @@ get_line(char **buffer, size_t *have, FILE *fp)
 	if (used + 2 > *have) {
 	    adjust_buffer(buffer, *have *= 2);
 	}
-	(*buffer)[used++] = ch;
+	(*buffer)[used++] = (char) ch;
 	if (ch == '\n')
 	    break;
     }
@@ -522,7 +549,9 @@ get_line(char **buffer, size_t *have, FILE *fp)
 static void
 do_file(FILE *fp)
 {
-    DATA dummy, *that = &dummy;
+    DATA dummy;
+    DATA *that = &dummy;
+    DATA *prev = 0;
     char *buffer = 0;
     char *b_fname = 0;
     char *b_temp1 = 0;
@@ -535,12 +564,14 @@ do_file(FILE *fp)
     int unified = 0;
     int old_unify = 0;
     int new_unify = 0;
+    int context = 1;
     char *s;
+#ifdef DEBUG
+    int line_no = 0;
+#endif
 
+    memset(&dummy, 0, sizeof(dummy));
     dummy.name = "";
-    dummy.ins = 0;
-    dummy.del = 0;
-    dummy.mod = 0;
 
     fixed_buffer(&buffer, fixed = length = BUFSIZ);
     fixed_buffer(&b_fname, length);
@@ -564,26 +595,43 @@ do_file(FILE *fp)
 	 * Trim trailing blanks (e.g., newline)
 	 */
 	for (s = buffer + strlen(buffer); s > buffer; s--) {
-	    if (isspace(s[-1]))
+	    if (isspace(UC(s[-1])))
 		s[-1] = EOS;
 	    else
 		break;
 	}
+	TRACE(("[%05d] %s\n", ++line_no, buffer));
 
 	/*
+	 * The lines identifying files in a context diff depend on how it was
+	 * invoked.  But after the header, each chunk begins with a line
+	 * containing 15 *'s.  Each chunk may contain a line-range with '***'
+	 * for the "before", and a line-range with '---' for the "after".  The
+	 * part of the chunk depicting the deletion may be absent, though the
+	 * edit line is present.
+	 *
 	 * The markers for unified diff are a little different from the normal
 	 * context-diff.  Also, the edit-lines in a unified diff won't have a
 	 * space in column 2.  Because of the missing space, we have to count
 	 * lines to ensure we do not confuse the marker lines.
 	 */
 	marker = -1;
-	if (match(buffer, "*** ")) {
+	if (that != &dummy && !strcmp(buffer, "***************")) {
+	    TRACE(("begin context chunk\n"));
+	    context = 2;
+	} else if (context == 2 && match(buffer, "*** ")) {
+	    context = 1;
+	} else if (context == 1 && match(buffer, "--- ")) {
+	    marker = 1;
+	    context = 0;
+	} else if (match(buffer, "*** ")) {
 	    marker = 0;
 	} else if ((old_unify + new_unify) == 0 && match(buffer, "--- ")) {
 	    marker = unified = 1;
 	} else if ((old_unify + new_unify) == 0 && match(buffer, "+++ ")) {
 	    marker = unified = 2;
-	} else if (unified == 2) {
+	} else if (unified == 2
+		   || ((old_unify + new_unify) == 0 && (*buffer == '@'))) {
 	    unified = 0;
 	    if (*buffer == '@') {
 		int old_base, new_base, old_size, new_size;
@@ -602,6 +650,29 @@ do_file(FILE *fp)
 		    unified = -1;
 		}
 	    }
+	} else if (unified == 1 && !context) {
+	    /*
+	     * If unified==1, we guessed we would find a "+++" line, but since
+	     * we are here, we did not find that.  The context check ensures
+	     * we do not mistake the "---" for a unified diff with that for
+	     * a context diff's "after" line-range.
+	     *
+	     * If we guessed wrong, then we probably found a data line with
+	     * "--" in the first two columns of the diff'd file.
+	     */
+	    unified = 0;
+	    TRACE(("Expected \"+++\" @%d:%s\n", __LINE__, buffer));
+	    if (prev != 0
+		&& prev != that
+		&& that->ins == 0
+		&& that->del == 0
+		&& strcmp(prev->name, that->name)) {
+		TRACE(("giveup on %ld/%ld %s\n", that->ins, that->del, that->name));
+		TRACE(("revert to %ld/%ld %s\n", prev->ins, prev->del, prev->name));
+		delink(that);
+		that = prev;
+		that->del += 1;
+	    }
 	} else if (old_unify + new_unify) {
 	    switch (*buffer) {
 	    case '-':
@@ -625,8 +696,15 @@ do_file(FILE *fp)
 	} else {
 	    unified = 0;
 	}
-	if (marker > 0)
+
+	/*
+	 * Override the beginning of the line to simplify the case statement
+	 * below.
+	 */
+	if (marker > 0) {
+	    TRACE(("@%d, marker=%d, override %s\n", __LINE__, marker, buffer));
 	    (void) strncpy(buffer, "***", 3);
+	}
 
 	/*
 	 * Use the first character of the input line to determine its
@@ -636,10 +714,10 @@ do_file(FILE *fp)
 	case 'O':		/* Only */
 	    if (match(buffer, "Only in ")) {
 		char *path = buffer + 8;
-		int found = FALSE;
+		int found = 0;
 		for (s = path; *s != EOS; s++) {
 		    if (match(s, ": ")) {
-			found = TRUE;
+			found = 1;
 			*s++ = PATHSEP;
 			while ((s[0] = s[1]) != EOS)
 			    s++;
@@ -729,6 +807,7 @@ do_file(FILE *fp)
 			&& !contain_any(b_fname, "*")
 			&& !edit_range(b_fname))
 		    ) {
+		    prev = that;
 		    s = do_merging(that, b_fname);
 		    that = new_data(s);
 		    ok = begin_data(that);
@@ -738,31 +817,26 @@ do_file(FILE *fp)
 	    break;
 
 	case '+':
-	    if (!unified && buffer[0] == buffer[1])
-		break;
 	    /* FALL-THRU */
 	case '>':
-	    if (!ok)
-		break;
-	    that->ins += 1;
+	    if (ok)
+		that->ins += 1;
 	    break;
 
 	case '-':
 	    if (!ok)
 		break;
-	    if (!unified && buffer[0] == buffer[1])
+	    if (!unified && !strcmp(buffer, "---"))
 		break;
 	    /* fall-thru */
 	case '<':
-	    if (!ok)
-		break;
-	    that->del += 1;
+	    if (ok)
+		that->del += 1;
 	    break;
 
 	case '!':
-	    if (!ok)
-		break;
-	    that->mod += 1;
+	    if (ok)
+		that->mod += 1;
 	    break;
 
 	    /* Expecting "Binary files XXX and YYY differ" */
@@ -926,7 +1000,7 @@ summarize(void)
 
 #ifdef HAVE_POPEN
 static char *
-is_compressed(char *name)
+is_compressed(const char *name)
 {
     char *verb = 0;
     char *result = 0;
@@ -940,7 +1014,7 @@ is_compressed(char *name)
 	verb = "bzip2 -dc %s";
     }
     if (verb != 0) {
-	result = (char *) malloc(strlen(verb) + len);
+	result = (char *) xmalloc(strlen(verb) + len);
 	sprintf(result, verb, name);
     }
     return result;
@@ -951,7 +1025,7 @@ is_compressed(char *name)
 static void
 usage(FILE *fp)
 {
-    static char *msg[] =
+    static const char *msg[] =
     {
 	"Usage: diffstat [options] [files]",
 	"",
@@ -978,6 +1052,26 @@ usage(FILE *fp)
 	fprintf(fp, "%s\n", msg[j]);
 }
 
+/* Wrapper around getopt that also parses "--help" and "--version".  
+ * argc, argv, opts, return value, and globals optarg, optind,
+ * opterr, and optopt are as in getopt().  help and version designate
+ * what should be returned if --help or --version are encountered. */
+static int
+getopt_helper(int argc, char *const argv[], const char *opts,
+	      int help, int version)
+{
+    if (optind < argc && argv[optind] != NULL) {
+	if (strcmp(argv[optind], "--help") == 0) {
+	    optind++;
+	    return help;
+	} else if (strcmp(argv[optind], "--version") == 0) {
+	    optind++;
+	    return version;
+	}
+    }
+    return getopt(argc, argv, opts);
+}
+
 int
 main(int argc, char *argv[])
 {
@@ -986,7 +1080,8 @@ main(int argc, char *argv[])
 
     max_width = 80;
 
-    while ((j = getopt(argc, argv, "ce:f:hkn:o:p:uvVw:")) != EOF) {
+    while ((j = getopt_helper(argc, argv, "ce:f:hkn:o:p:uvVw:", 'h', 'V'))
+	   != -1) {
 	switch (j) {
 	case 'c':
 	    comment_opt = "#";
@@ -1021,7 +1116,9 @@ main(int argc, char *argv[])
 	    verbose = 1;
 	    break;
 	case 'V':
+#ifndef	NO_IDENT
 	    if (!sscanf(Id, "%*s %*s %s", version))
+#endif
 		(void) strcpy(version, "?");
 	    printf("diffstat version %s\n", version);
 	    return (EXIT_SUCCESS);
@@ -1050,11 +1147,11 @@ main(int argc, char *argv[])
 		    }
 		    do_file(fp);
 		    (void) pclose(fp);
-		    free(command);
 		}
+		free(command);
 	    } else
 #endif
-	    if ((fp = fopen(name, "r")) != 0) {
+	    if ((fp = fopen(name, "rb")) != 0) {
 		if (show_progress) {
 		    (void) fprintf(stderr, "%s\n", name);
 		    (void) fflush(stderr);
